from datetime import datetime

import pytest
from git import GitCommandError

from legendmeta import LegendMetadata
from legendmeta.jsondb import AttrsDict

pytestmark = pytest.mark.xfail(run=True, reason="requires access to legend-metadata")


@pytest.fixture(scope="module")
def metadb():
    mdata = LegendMetadata()
    mdata.checkout("98c00f0")
    return mdata


def test_get_file(metadb):
    metadb["hardware/detectors/germanium/diodes/B00000A.json"]


def test_get_directory(metadb):
    metadb["hardware"]
    metadb.hardware


def test_file_not_found(metadb):
    with pytest.raises(FileNotFoundError):
        metadb["non-existing-file.ext"]


def test_git_ref_not_found(metadb):
    with pytest.raises(GitCommandError):
        metadb.checkout("non-existent-ref")


def test_nested_get(metadb):
    assert (
        metadb["hardware"]["detectors"]["germanium"]["diodes"]["B00000A"]["name"]
        == "B00000A"
    )
    assert metadb.hardware.detectors.germanium.diodes.B00000A.name == "B00000A"


def test_chmap_remapping(metadb):
    assert (
        "daq"
        in metadb.hardware.configuration.channelmaps.on(datetime.now()).map(
            "daq.rawid"
        )[1080000]
    )

    assert "daq" in metadb.channelmap().map("daq.rawid")[1080000]


def test_channelmap(metadb):
    channel = metadb.channelmap(on=datetime.now()).V02162B
    assert isinstance(channel, AttrsDict)
    assert "geometry" in channel
    assert hasattr(channel, "geometry")
    assert "analysis" in channel



# Copyright (C) 2022 Luigi Pertoldi <gipert@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import json
import logging
import os
import re
from datetime import datetime
from pathlib import Path
from typing import Any, Iterator

from legendmeta.catalog import Catalog, Props

log = logging.getLogger(__name__)


class AttrsDict(dict):
    """Access dictionary items as attributes.

    Examples
    --------
    >>> d = AttrsDict({"key1": {"key2": 1}})
    >>> d.key1.key2 # == 1
    >>> d1 = AttrsDict()
    >>> d1["a"] = 1
    >>> d1.a # == 1
    """

    def __init__(self, value: dict = None) -> None:
        if value is None:
            super().__init__()
        # can only be initialized with a dict
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise TypeError("expected dict")

        # attribute that holds cached remappings -- see map()
        super().__setattr__("_cached_remaps", {})

    def __setitem__(self, key: str | int | float, value: Any) -> Any:
        # convert dicts to AttrsDicts
        if not isinstance(value, AttrsDict):
            if isinstance(value, dict):
                value = AttrsDict(value)  # this should make it recursive
            # recurse lists
            elif isinstance(value, list):
                for i, el in enumerate(value):
                    if isinstance(el, dict):
                        value[i] = AttrsDict(el)  # this should make it recursive

        super().__setitem__(key, value)

        # if the key is a valid attribute name, create a new attribute
        if isinstance(key, str) and key.isidentifier():
            super().__setattr__(key, value)

        # reset special _cached_remaps private attribute -- see map()
        super().__setattr__("_cached_remaps", {})

    __setattr__ = __setitem__

    def __getattr__(self, name: str) -> Any:
        try:
            super().__getattr__(name)
        except AttributeError:
            raise AttributeError(f"dictionary does not contain a '{name}' key")

    def map(self, label: str, unique: bool = True) -> AttrsDict:
        """Remap dictionary according to an alternative unique label.

        Loop over keys in the first level and search for key named `label` in
        their values. If `label` is found and its value `newid` is unique,
        create a mapping between `newid` and the first-level dictionary `obj`.
        If `label` is of the form ``key.label``, ``label`` will be searched in
        a dictionary keyed by ``key``. If the label is unique a dictionary of
        dictionaries will be returned, if not unique and `unique` is false, a
        dictionary will be returned where each entry is a dictionary of
        dictionaries keyed by an arbitrary integer.

        Parameters
        ----------
        label
            name (key) at which the new label can be found. If nested in
            dictionaries, use ``.`` to separate levels, e.g.
            ``level1.level2.label``.
        unique
            bool specifying whether only unique keys are allowed. If true
            will raise an error if the specified key is not unique.

        Examples
        --------
        >>> d = AttrsDict({
        ...   "a": {
        ...     "id": 1,
        ...     "group": {
        ...       "id": 3,
        ...     },
        ...     "data": "x"
        ...   },
        ...   "b": {
        ...     "id": 2,
        ...     "group": {
        ...       "id": 4,
        ...     },
        ...     "data": "y"
        ...   },
        ... })
        >>> d.map("id")[1].data == "x"
        True
        >>> d.map("group.id")[4].data == "y"
        True

        Note
        ----
        No copy is performed, the returned dictionary is made of references to
        the original objects.

        Warning
        -------
        The result is cached internally for fast access after the first call.
        If the dictionary is modified, the cache gets cleared.
        """
        # if this is a second call, return the cached result
        if label in self._cached_remaps:
            return self._cached_remaps[label]

        splitk = label.split(".")
        newmap = AttrsDict()
        unique_tracker = True
        # loop over values in the first level
        for v in self.values():
            # find the (nested) label value
            newid = v
            try:
                for k in splitk:
                    newid = newid[k]
            # just skip if the label is not there
            except (KeyError, TypeError, FileNotFoundError):
                continue

            if not isinstance(newid, (int, float, str)):
                raise RuntimeError(f"'{label}' values are not all numbers or strings")
            # complain if a label with the same value was already found
            if newid in newmap:
                newkey = sorted(newmap[newid].keys())[-1] + 1
                newmap[newid].update({newkey: v})
                unique_tracker = False
            else:
                # add an item to the new dict with key equal to the value of the label
                newmap[newid] = {0: v}

        if unique is True and unique_tracker is False:
            raise RuntimeError(f"'{label}' values are not unique")

        if unique_tracker is True:
            newmap = AttrsDict({entry: newmap[entry][0] for entry in newmap})

        if not newmap:
            raise ValueError(f"could not find '{label}' anywhere in the dictionary")

        # cache it
        self._cached_remaps[label] = newmap
        return newmap

    # d |= other_d should still produce a valid AttrsDict
    def __ior__(self, other: dict | AttrsDict) -> AttrsDict:
        return AttrsDict(super().__ior__(other))

    # d1 | d2 should still produce a valid AttrsDict
    def __or__(self, other: dict | AttrsDict) -> AttrsDict:
        return AttrsDict(super().__or__(other))


class JsonDB:
    """Bare-bones JSON database.

    The database is represented on disk by a collection of JSON files
    arbitrarily scattered in a filesystem. Subdirectories are also
    :class:`.JsonDB` objects. In memory, the database is represented as an
    :class:`AttrsDict`.

    Note
    ----
    For large databases, a basic "lazy" mode is available. In this case, no
    global scan of the filesystem is performed at initialization time. Once a
    file is queried, it is also cached in the internal store for faster access.
    A call to :meth:`scan` is needed before iterating on the database files.

    Examples
    --------
    >>> from legendmeta.jsondb import JsonDB
    >>> jdb = JsonDB("path/to/dir")
    >>> jdb["file1.json"]  # is a dict
    >>> jdb["file1"]  # also works
    >>> jdb["dir1"]  # JsonDB instance
    >>> jdb["dir1"]["file1"]  # nested JSON file
    >>> jdb["dir1/file1"]  # also works
    >>> jdb.dir1.file # keys can be accessed as attributes
    """

    def __init__(self, path: str | Path, lazy: bool = False) -> None:
        """Construct a :class:`.JsonDB` object.

        Parameters
        ----------
        path
            path to the directory containing the database.
        lazy
            whether a database scan should be performed at initialization time.
        """
        self.path: Path = Path(path).expanduser().resolve()
        if not self.path.is_dir():
            raise ValueError("input path is not a valid directory")

        self._store = AttrsDict()

        if not lazy:
            self.scan()

    def scan(self, subdir: str = ".") -> None:
        """Populate the database by walking the filesystem.

        Parameters
        ----------
        subdir
            restrict scan to path relative to the database location.
        """
        for j in self.path.rglob(f"{subdir}/*.json"):
            try:
                self[j]
            except (json.JSONDecodeError, ValueError) as e:
                log.warning(f"could not scan file {j}")
                log.warning(f"reason: {e}")

    def keys(self) -> list[str]:
        return self._store.keys()

    def items(self) -> Iterator[(str, JsonDB | AttrsDict | list)]:
        return self._store.items()

    def on(
        self, timestamp: str | datetime, pattern: str = None, system: str = "all"
    ) -> AttrsDict | list:
        """Query database in `time[, file pattern, system]`.

        A (only one) valid ``validity.jsonl`` file must exist in the directory
        to specify a validity mapping. This functionality relies on the
        :class:`.catalog.Catalog` class.

        The JSONL specification is documented at `this link
        <https://legend-exp.github.io/legend-data-format-specs/dev/metadata/#Specifying-metadata-validity-in-time-(and-system)>`_.

        The special ``$_`` string is expanded to the directory containing the
        JSON files.

        Parameters
        ----------
        timestamp
            a :class:`~datetime.datetime` object or a string matching the
            pattern ``YYYYmmddTHHMMSSZ``.
        pattern
            query by filename pattern.
        system: 'all', 'phy', 'cal', 'lar', ...
            query only a data taking "system".
        """
        jsonl = os.path.join(self.path, "validity.jsonl")
        if not os.path.isfile(jsonl):
            raise RuntimeError(f"no validity.jsonl file found in {self.path}")

        file_list = Catalog.get_files(jsonl, timestamp, system)
        # select only files matching pattern if specified
        if pattern is not None:
            c = re.compile(pattern)
            out_files = []
            for file in file_list:
                if c.match(file):
                    out_files.append(file)
            files = out_files
        else:
            files = file_list

        # read files in and combine as necessary
        db_ptr = self
        if isinstance(files, list):
            result = AttrsDict()
            for file in files:
                fp = self.path.rglob(file)
                fp = [i for i in fp][0]
                # TODO: what does this do exactly?
                Props.add_to(result, db_ptr[fp])
            db_ptr = result
        else:
            fp = self.path.rglob(file)
            fp = [i for i in fp][0]
            db_ptr = db_ptr[fp]
        Props.subst_vars(db_ptr, var_values={"_": self.path})
        return db_ptr

    def map(self, label: str, unique: bool = True) -> AttrsDict:
        """Remap dictionary according to a second unique `key`.

        See Also
        --------
        AttrsDict.map

        Warning
        -------
        If the database is lazy, call :meth:`.scan` in advance to populate it
        (or a subdirectory).
        """
        return self._store.map(label, unique=unique)

    def __getitem__(self, item: str | Path) -> JsonDB | AttrsDict | list:
        """Access files or directories in the database."""
        # resolve relative paths / links, but keep it relative to self.path
        item = Path(self.path / item).expanduser().resolve().relative_to(self.path)

        # now call this very function recursively to walk the directories to the file
        db_ptr = self
        for d in item.parts[0:-1]:
            db_ptr = db_ptr[d]

        # item_id should not contain any / at this point
        # store JSON file names without extension
        item_id = item.stem
        # skip if object is already in the store
        if item_id not in db_ptr._store:
            obj = db_ptr.path / item.name
            # if directory, construct another JsonDB object
            if obj.is_dir():
                db_ptr._store[item_id] = JsonDB(obj)
            else:
                # try to attach .json extension if file cannot be found
                if not obj.is_file():
                    obj = Path(str(obj) + ".json")

                # if it's a valid JSON file, construct an AttrsDict object
                if obj.is_file():
                    with obj.open() as f:
                        loaded = json.load(f)
                        if isinstance(loaded, dict):
                            loaded = AttrsDict(loaded)
                            Props.subst_vars(loaded, var_values={"_": self.path})
                        else:  # must be a list, check if there are dicts inside to convert
                            for i, el in enumerate(loaded):
                                if isinstance(el, dict):
                                    loaded[i] = AttrsDict(el)
                                    Props.subst_vars(
                                        loaded[i], var_values={"_": self.path}
                                    )

                        db_ptr._store[item_id] = loaded
                else:
                    raise FileNotFoundError(
                        f"{str(obj).replace('.json.json', '.json')} is not a valid file or directory"
                    )

            # set also an attribute, if possible
            if item_id.isidentifier():
                self.__setattr__(item_id, db_ptr._store[item_id])

        return db_ptr._store[item_id]

    def __getattr__(self, name: str) -> JsonDB | AttrsDict | list:
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"database does not contain '{name}'")

    # NOTE: self cannot stay a JsonDB, since the class is characterized by a
    # (unique) root directory. What would be the root directory of the merged
    # JsonDB?
    def __ior__(self, other: JsonDB) -> AttrsDict:
        raise TypeError("cannot merge JsonDB in-place")

    # NOTE: returning a JsonDB does not make much sense, see above
    def __or__(self, other: JsonDB) -> AttrsDict:
        if isinstance(other, JsonDB):
            return self._store | other._store
        else:
            return self._store | other

    def __len__(self) -> int:
        return len(self._store)

    def __iter__(self) -> Iterator:
        return iter(self._store)

    def __str__(self) -> str:
        return str(self._store)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{str(self.path)}')"



# Copyright (C) 2022 Luigi Pertoldi <gipert@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import os
from datetime import datetime

import pandas
import sqlalchemy as db

from .core import AttrsDict
from .scdb_tables import (
    DiodeConfMon,
    DiodeInfo,
    DiodeSnap,
    MuonConfMon,
    MuonInfo,
    MuonSnap,
    SiPMConfMon,
    SiPMInfo,
    SiPMSnap,
)

log = logging.getLogger(__name__)


class LegendSlowControlDB:
    """A class for interacting with the LEGEND Slow Control database.

    The LEGEND Slow Control system relies on a `PostgreSQL
    <https://www.postgresql.org/docs/current/index.html>`_ database living on
    ``legend-sc.lngs.infn.it``. The aim of the :class:`LegendSlowControlDB`
    class is to simplify access to the database from Python.
    """

    def __init__(self, connect=False) -> None:
        self.connection: db.engine.base.Connection = None
        self.session: db.orm.Session = None

        if connect:
            self.connect()

    def connect(
        self, host: str = "localhost", port: int = 5432, password: str | None = None
    ) -> None:
        """Establish a connection to the database.

        Authentication is attempted with the read-only user ``scuser`` on a
        database named ``scdb``.

        Parameters
        ----------
        host
            database host. Can be a hostname (``localhost``,
            ``legend-sc.lngs.infn.it``, etc.) or an IP address.
        port
            port through which the database should be contacted.
        password
            password for user ``scuser`` of the ``scdb`` database. May be found
            on LEGEND's internal documentation (e.g. the Wiki web pages). If
            ``None``, uses the value of the ``$LEGEND_SCDB_PW`` shell variable.

        Examples
        --------
        If the Slow Control database connection is forwarded to a local machine
        (port 6942) (through e.g. an SSH tunnel), use:

        >>> scdb = LegendSlowControlDB()
        >>> scdb.connect("localhost", port=6942, password="···")

        Alternatively to giving the password to ``connect()``, it can be stored
        in the ``$LEGEND_SCDB_PW`` shell variable (in e.g. ``.bashrc``):

        .. code-block:: bash
           :caption: ``~/.bashrc``

           export LEGEND_SCDB_PW="···"

        Then:

        >>> scdb.connect("localhost", port=6942)
        """
        if password is None:
            password = os.getenv("LEGEND_SCDB_PW")

        if password is None:
            raise ValueError("must supply the database password")

        if self.connection is not None and not self.connection.closed:
            self.disconnect()

        self.connection = db.create_engine(
            f"postgresql://scuser:{password}@{host}:{port}/scdb"
        ).connect()

    def disconnect(self) -> None:
        """Disconnect from the database."""
        self.connection.close()

    def make_session(self) -> db.orm.Session:
        """Open and return a new  :class:`~sqlalchemy.orm.Session` object for executing database operations.

        Examples
        --------
        >>> import sqlalchemy as sql
        >>> from legendmeta.slowcontrol import DiodeSnap
        >>> session = scdb.make_session()
        >>> result = session.execute(sql.select(DiodeSnap.channel, DiodeSnap.imon).limit(3))
        >>> result.all()
        [(2, 0.0007), (1, 0.0001), (5, 5e-05)]

        See Also
        --------
        `SQLAlchemy documentation <https://www.sqlalchemy.org/>`_
        """
        if self.session:
            log.warning(
                "A session seems to be already open and available "
                "at the .session attribute of this class object. "
                "You might want to use that one."
            )

        return db.orm.Session(self.connection)

    def get_tables(self) -> list[str]:
        """Get tables available in the database."""
        return db.inspect(self.connection.engine).get_table_names()

    def get_columns(self, table: str) -> list[str]:
        """Get columns available on `table` in the database."""
        return db.inspect(self.connection.engine).get_columns(table)

    def dataframe(self, expr: str | db.sql.Select) -> pandas.DataFrame:
        """Query the database and return a dataframe holding the result.

        Parameters
        ----------
        expr
            SQL table name, select SQL command text or SQLAlchemy selectable
            object.

        Examples
        --------
        SQL select syntax text or table name:

        >>> scdb.dataframe("SELECT channel, vmon FROM diode_snap LIMIT 3")
           channel    vmon
        0        2  2250.0
        1        1  3899.4
        2        5  1120.2

        >>> scdb.dataframe("diode_conf_mon")
              confid  crate  slot  channel    vset  iset  rup  rdown  trip  vmax pwkill pwon                    tstamp
        0         15      0     0        0  4000.0   6.0   10      5  10.0  6000   KILL  Dis 2022-10-07 13:49:56+00:00
        1         15      0     0        1  4300.0   6.0   10      5  10.0  6000   KILL  Dis 2022-10-07 13:49:56+00:00
        2         15      0     0        2  4200.0   6.0   10      5  10.0  6000   KILL  Dis 2022-10-07 13:49:56+00:00
        ...

        :class:`sqlalchemy.sql.Select` object:

        >>> import sqlalchemy as sql
        >>> from legendmeta.slowcontrol import DiodeSnap
        >>> scdb.dataframe(sql.select(DiodeSnap.channel, DiodeSnap.vmon).limit(3))
           channel    vmon
        0        2  2250.0
        1        1  3899.4
        2        5  1120.2

        See Also
        --------
        pandas.read_sql
        """
        try:
            try:
                return pandas.read_sql(expr, self.connection)
            except db.exc.ObjectNotExecutableError:
                return pandas.read_sql(db.text(expr), self.connection)
        # try rolling back transaction if any exception occurs
        except Exception:
            self.connection.rollback()
            raise

    def status(
        self, channel: dict, on: str | datetime = None, system: str = None
    ) -> dict:
        """Query status of a LEGEND DAQ channel.

        Collect all the relevant information about the status of a channel at a
        certain time from the Slow Control database, based on the channel type
        (i.e. germanium detector, SiPM or PMT).

        Parameters
        ----------
        channel
            this dictionary must contain information about the channel hardware
            configuration. Typically a LEGEND channel map entry (obtainable, for
            example, with :meth:`.core.LegendMetadata.channelmap`).
        on
            time at which the status is requested.
        system: "geds", "spms", "pmts", ...
            system the channel belong to. this information is used to ask the
            Slow Control database the right questions. If ``None`` will try to
            determine it from the available metadata.

        Examples
        --------
        >>> ts = datetime(...) # or LEGEND cycle timestamp
        >>> my_bege = lmeta.channelmap(on=ts).B00089B
        >>> scdb.status(my_bege, on=ts)
        {'vmon': 3399.9,
         'imon': 0.00015,
         'status': 1,
         'vset': 3400.0,
         'iset': 6.0,
         ...

        Warning
        -------
        This class method assumes a certain structure for legend-metadata.
        Might stop working if that structure is altered.
        """
        if not on:
            on = datetime.now()

        # sanity checks
        if isinstance(on, str):
            on = datetime.strptime(on, "%Y%m%dT%H%M%SZ")
        elif not isinstance(on, datetime):
            raise ValueError("Bad input timestamp format")

        if not isinstance(channel, dict):
            raise ValueError("Bad channel format: dict expected")

        if not system:
            system = channel.system

        # prepare environment to perform query
        if not self.session:
            self.session = self.make_session()

        output = AttrsDict()
        if system == "geds":
            for tbl in [DiodeInfo, DiodeSnap, DiodeConfMon]:
                stmt = (
                    db.select(tbl)
                    .where(tbl.slot == channel.voltage.card.id)
                    .where(tbl.channel == channel.voltage.channel)
                    .order_by(tbl.tstamp.desc())
                    .where(tbl.tstamp <= on)
                    .limit(1)
                )

                result = self.session.execute(stmt).first()

                if not result:
                    log.warning(
                        f"Query on table '{tbl.__tablename__}' did not produce any result"
                    )
                    continue

                output |= result[0].asdict()

        elif system == "spms":
            for tbl in [SiPMInfo, SiPMSnap, SiPMConfMon]:
                stmt = (
                    db.select(tbl)
                    .where(tbl.board == channel.electronics.card.id)
                    .where(tbl.channel == channel.electronics.channel)
                    .order_by(tbl.tstamp.desc())
                    .where(tbl.tstamp <= on)
                    .limit(1)
                )

                result = self.session.execute(stmt).first()

                if not result:
                    log.warning(
                        f"Query on table '{tbl.__tablename__}' did not produce any result"
                    )
                    continue

                output |= result[0].asdict()

        # NOTE: untested
        elif system == "pmts":
            for tbl in [MuonInfo, MuonSnap, MuonConfMon]:
                stmt = (
                    db.select(tbl)
                    .where(tbl.slot == channel.voltage.card.id)
                    .where(tbl.channel == channel.voltage.channel)
                    .order_by(tbl.tstamp.desc())
                    .where(tbl.tstamp <= on)
                    .limit(1)
                )

                result = self.session.execute(stmt).first()

                if not result:
                    log.warning(
                        f"Query on table '{tbl.__tablename__}' did not produce any result"
                    )
                    continue

                output |= result[0].asdict()
        else:
            raise NotImplementedError(f"System '{system}' not (yet) supported")

        if not output:
            raise RuntimeError("Could not obtain any information about the channel")

        return output



# Copyright (C) 2022 Luigi Pertoldi <gipert@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime

import sqlalchemy as db
from sqlalchemy.orm import DeclarativeBase, Mapped

from .jsondb import AttrsDict


class Base(DeclarativeBase):
    pass


@dataclass
class DiodeSnap(Base):
    """Monitored parameters of HPGe detectors."""

    __tablename__ = "diode_snap"

    crate: Mapped[int]
    slot: Mapped[int]
    channel: Mapped[int]
    vmon: Mapped[float]
    imon: Mapped[float]
    status: Mapped[int]
    almask: Mapped[int]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict({"vmon": self.vmon, "imon": self.imon, "status": self.status})


@dataclass
class DiodeConfMon(Base):
    """Configuration parameters of HPGe detectors."""

    __tablename__ = "diode_conf_mon"

    confid: Mapped[int]
    crate: Mapped[int]
    slot: Mapped[int]
    channel: Mapped[int]
    vset: Mapped[float]
    iset: Mapped[float]
    rup: Mapped[int]
    rdown: Mapped[int]
    trip: Mapped[float]
    vmax: Mapped[int]
    pwkill: Mapped[str]
    pwon: Mapped[str]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict(
            {
                "vset": self.vset,
                "iset": self.iset,
                "rup": self.rup,
                "rdown": self.rdown,
                "trip": self.trip,
                "vmax": self.vmax,
                "pwkill": self.pwkill,
                "pwon": self.pwon,
            }
        )


@dataclass
class DiodeInfo(Base):
    """Static information about HPGe detectors."""

    __tablename__ = "diode_info"

    crate: Mapped[int]
    slot: Mapped[int]
    channel: Mapped[int]
    group: Mapped[str]
    label: Mapped[str]
    status: Mapped[int]
    itol: Mapped[float]
    vtol: Mapped[float]
    iupd: Mapped[float]
    vupd: Mapped[float]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict({"group": self.group, "label": self.label})


@dataclass
class SiPMSnap(Base):
    """Monitored parameters of SiPMs from the LAr instrumentation."""

    __tablename__ = "sipm_snap"

    board: Mapped[int]
    channel: Mapped[int]
    vmon: Mapped[float]
    imon: Mapped[float]
    status: Mapped[int]
    almask: Mapped[int]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict({"vmon": self.vmon, "imon": self.imon, "status": self.status})


class SiPMConfMon(Base):
    """Configuration parameters of SiPMs from the LAr instrumentation."""

    __tablename__ = "sipm_conf_mon"

    confid: Mapped[int]
    board: Mapped[int]
    channel: Mapped[int]
    vset: Mapped[float]
    iset: Mapped[float]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict(
            {
                "vset": self.vset,
                "iset": self.iset,
            }
        )


@dataclass
class SiPMInfo(Base):
    """Static information about SiPMs from the LAr instrumentation."""

    __tablename__ = "sipm_info"

    board: Mapped[int]
    channel: Mapped[int]
    group: Mapped[str]
    label: Mapped[str]
    status: Mapped[int]
    itol: Mapped[float]
    vtol: Mapped[float]
    iupd: Mapped[float]
    vupd: Mapped[float]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict({"group": self.group, "label": self.label})


@dataclass
class MuonSnap(Base):
    """Monitored parameters of PMTs from the muon veto."""

    __tablename__ = "muon_snap"

    crate: Mapped[int]
    slot: Mapped[int]
    channel: Mapped[int]
    vmon: Mapped[float]
    imon: Mapped[float]
    status: Mapped[int]
    almask: Mapped[int]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict({"vmon": self.vmon, "imon": self.imon, "status": self.status})


@dataclass
class MuonConfMon(Base):
    """Configuration parameters of PMTs from the muon veto."""

    __tablename__ = "muon_conf_mon"

    confid: Mapped[int]
    crate: Mapped[int]
    slot: Mapped[int]
    channel: Mapped[int]
    vset: Mapped[float]
    iset: Mapped[float]
    rup: Mapped[int]
    rdown: Mapped[int]
    trip: Mapped[float]
    vmax: Mapped[int]
    pwkill: Mapped[str]
    pwon: Mapped[str]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict(
            {
                "vset": self.vset,
                "iset": self.iset,
                "rup": self.rup,
                "rdown": self.rdown,
                "trip": self.trip,
                "vmax": self.vmax,
                "pwkill": self.pwkill,
                "pwon": self.pwon,
            }
        )


@dataclass
class MuonInfo(Base):
    """Static information about PMTs from the muon veto."""

    __tablename__ = "muon_info"

    crate: Mapped[int]
    slot: Mapped[int]
    channel: Mapped[int]
    group: Mapped[str]
    label: Mapped[str]
    status: Mapped[int]
    itol: Mapped[float]
    vtol: Mapped[float]
    iupd: Mapped[float]
    vupd: Mapped[float]
    tstamp: Mapped[datetime] = db.orm.mapped_column(primary_key=True)

    def asdict(self) -> AttrsDict:
        return AttrsDict({"group": self.group, "label": self.label})



# Copyright (C) 2015 Oliver Schulz <oschulz@mpp.mpg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import bisect
import collections
import copy
import json
import os
import types
from collections import namedtuple
from datetime import datetime
from string import Template


def to_datetime(value):
    """Convert a LEGEND timestamp (or key) to :class:`datetime.datetime`."""
    return datetime.strptime(value, "%Y%m%dT%H%M%SZ")


def unix_time(value):
    if isinstance(value, str):
        return datetime.timestamp(datetime.strptime(value, "%Y%m%dT%H%M%SZ"))
    elif isinstance(value, datetime):
        return datetime.timestamp(value)
    else:
        raise ValueError(f"Can't convert type {type(value)} to unix time")


class PropsStream:
    @staticmethod
    def get(value):
        if isinstance(value, str):
            return PropsStream.read_from(value)
        elif isinstance(value, collections.abc.Sequence) or isinstance(
            value, types.GeneratorType
        ):
            return value
        else:
            raise ValueError(f"Can't get PropsStream from value of type {type(value)}")

    @staticmethod
    def read_from(file_name):
        with open(file_name) as file:
            for json_str in file:
                yield json.loads(json_str)


class Catalog(namedtuple("Catalog", ["entries"])):
    """Implementation of the `JSONL metadata validity specification <https://legend-exp.github.io/legend-data-format-specs/dev/metadata/#Specifying-metadata-validity-in-time-(and-system)>`_."""

    __slots__ = ()

    class Entry(namedtuple("Entry", ["valid_from", "file"])):
        __slots__ = ()

    @staticmethod
    def get(value):
        if isinstance(value, Catalog):
            return value
        if isinstance(value, str):
            return Catalog.read_from(value)
        else:
            raise ValueError(f"Can't get Catalog from value of type {type(value)}")

    @staticmethod
    def read_from(file_name):
        entries = {}

        for props in PropsStream.get(file_name):
            timestamp = props["valid_from"]
            if props.get("category") is None:
                system = "all"
            else:
                system = props["category"]
            file_key = props["apply"]
            if system not in entries:
                entries[system] = []
            entries[system].append(Catalog.Entry(unix_time(timestamp), file_key))

        for system in entries:
            entries[system] = sorted(
                entries[system], key=lambda entry: entry.valid_from
            )
        return Catalog(entries)

    def valid_for(self, timestamp, system="all", allow_none=False):
        if system in self.entries:
            valid_from = [entry.valid_from for entry in self.entries[system]]
            pos = bisect.bisect_right(valid_from, unix_time(timestamp))
            if pos > 0:
                return self.entries[system][pos - 1].file
            elif system != "all":
                return self.valid_for(timestamp, system="all", allow_none=allow_none)
            else:
                if allow_none:
                    return None
                else:
                    raise RuntimeError(
                        f"No valid entries found for timestamp: {timestamp}, system: {system}"
                    )
        elif system != "all":
            return self.valid_for(timestamp, system="all", allow_none=allow_none)
        else:
            if allow_none:
                return None
            else:
                raise RuntimeError(f"No entries found for system: {system}")

    @staticmethod
    def get_files(catalog_file, timestamp, category="all"):
        catalog = Catalog.read_from(catalog_file)
        return Catalog.valid_for(catalog, timestamp, category)


class Props:
    @staticmethod
    def read_from(sources, subst_pathvar=False, trim_null=False):
        def read_impl(sources):
            if isinstance(sources, str):
                file_name = sources
                with open(file_name) as file:
                    result = json.load(file)
                    if subst_pathvar:
                        Props.subst_vars(
                            result,
                            var_values={"_": os.path.dirname(file_name)},
                            ignore_missing=True,
                        )
                    return result

            elif isinstance(sources, list):
                result = {}
                for p in map(read_impl, sources):
                    Props.add_to(result, p)
                return result
            else:
                raise ValueError(
                    f"Can't run Props.read_from on sources-value of type {type(sources)}"
                )

        result = read_impl(sources)
        if trim_null:
            Props.trim_null(result)
        return result

    @staticmethod
    def write_to(file_name, obj, pretty=False):
        separators = None if pretty else (",", ":")
        indent = 2 if pretty else None
        with open(file_name, "w") as file:
            json.dump(obj, file, indent=indent, separators=separators)
            file.write("\n")

    @staticmethod
    def add_to(props_a, props_b):
        a = props_a
        b = props_b

        for key in b:
            if key in a:
                if isinstance(a[key], dict) and isinstance(b[key], dict):
                    Props.add_to(a[key], b[key])
                elif a[key] != b[key]:
                    a[key] = copy.copy(b[key])
            else:
                a[key] = copy.copy(b[key])

    @staticmethod
    def trim_null(props_a):
        a = props_a

        for key in list(a.keys()):
            if isinstance(a[key], dict):
                Props.trim_null(a[key])
            elif a[key] is None:
                del a[key]

    @staticmethod
    def subst_vars(props, var_values=None, ignore_missing=False):
        if not var_values:
            var_values = {}

        for key in props:
            value = props[key]
            if isinstance(value, str) and "$" in value:
                new_value = None
                if ignore_missing:
                    new_value = Template(value).safe_substitute(var_values)
                else:
                    new_value = Template(value).substitute(var_values)

                if new_value != value:
                    props[key] = new_value
            elif isinstance(value, dict):
                Props.subst_vars(value, var_values, ignore_missing)



# Copyright (C) 2022 Luigi Pertoldi <gipert@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A package to access `legend-metadata <https://github.com/legend-exp/legend-metadata>`_ in Python."""

from legendmeta._version import version as __version__
from legendmeta.catalog import to_datetime
from legendmeta.core import LegendMetadata
from legendmeta.jsondb import JsonDB
from legendmeta.slowcontrol import LegendSlowControlDB

__all__ = [
    "__version__",
    "LegendMetadata",
    "LegendSlowControlDB",
    "JsonDB",
    "to_datetime",
]



from datetime import datetime, timezone
from pathlib import Path

import pytest

from legendmeta.jsondb import AttrsDict, JsonDB, Props

testdb = Path(__file__).parent / "testdb"


def test_props():
    # test read_from
    test_dict = Props.read_from(str(Path(__file__).parent / "testdb/file2.json"))
    assert test_dict["data"] == 2

    # test subst_vars
    Props.subst_vars(test_dict, var_values={"_": str(Path(__file__).parent / "testdb")})
    assert test_dict["filepath"] == str(
        Path(__file__).parent / "testdb/dir1/file3.json"
    )

    test_dict2 = Props.read_from(str(Path(__file__).parent / "testdb/file3.json"))

    # test add_to
    Props.add_to(test_dict, test_dict2)
    assert test_dict["data"] == 3

    # test trim null
    Props.trim_null(test_dict)
    with pytest.raises(KeyError):
        test_dict["null_key"]

    test_dict = Props.read_from(
        [
            str(Path(__file__).parent / "testdb/file2.json"),
            str(Path(__file__).parent / "testdb/file3.json"),
        ],
        subst_pathvar=True,
        trim_null=True,
    )
    assert test_dict["data"] == 3
    assert test_dict["filepath"] == str(
        Path(__file__).parent / "testdb/dir1/file3.json"
    )
    with pytest.raises(KeyError):
        test_dict["null_key"]


def test_access():
    jdb = JsonDB(testdb)
    assert isinstance(jdb["file1.json"], AttrsDict)
    assert isinstance(jdb["file2.json"], AttrsDict)
    assert isinstance(jdb["file1"], AttrsDict)
    assert isinstance(jdb["dir1"], JsonDB)
    assert isinstance(jdb["dir1"]["file3.json"], AttrsDict)
    assert isinstance(jdb["dir1"]["file3"], AttrsDict)
    assert isinstance(jdb["dir1/file3.json"], AttrsDict)
    assert isinstance(jdb["dir1"]["dir2"], JsonDB)
    assert isinstance(jdb["dir1"]["dir2"]["file4.json"], AttrsDict)
    assert isinstance(jdb["dir1/dir2/file4.json"], AttrsDict)
    assert jdb["file1.json"]["data"] == 1
    assert isinstance(jdb["file1"]["group"], AttrsDict)

    assert isinstance(jdb.file1, AttrsDict)
    assert isinstance(jdb.file1.group, AttrsDict)
    assert isinstance(jdb.dir1, JsonDB)
    assert isinstance(jdb.dir1.file3, AttrsDict)
    assert jdb.file1.data == 1
    assert jdb.file2.data == 2
    assert jdb.dir1.file3.data == 1
    assert jdb.file1.group.data1 == 1

    assert isinstance(jdb.arrays, list)
    assert jdb.arrays[0] == 0
    assert isinstance(jdb.arrays[1], AttrsDict)
    assert jdb.arrays[1].data == 1
    assert isinstance(jdb.arrays[1].array, list)
    assert isinstance(jdb.arrays[1].array[1], AttrsDict)
    assert jdb.arrays[1].array[0] == 1
    assert jdb.arrays[1].array[1].data == 2

    assert jdb.file2.filepath == str(Path(__file__).parent / "testdb/dir1/file3.json")

    with pytest.raises(ValueError):
        JsonDB("non-existent-db")
    with pytest.raises(FileNotFoundError):
        jdb["non-existent-file"]
    with pytest.raises(FileNotFoundError):
        jdb.non_existent_file

    with pytest.raises(AttributeError):
        jdb.dir1.file3.non_existent_key


def test_keys():
    jdb = JsonDB(testdb)
    assert sorted(jdb.keys()) == ["arrays", "dir1", "dir2", "file1", "file2", "file3"]
    assert sorted(jdb.dir1.keys()) == ["dir2", "file3", "file5"]


def test_items():
    jdb = JsonDB(testdb)
    items = sorted(jdb.items())
    assert items[0][0] == "arrays"
    assert isinstance(items[0][1], list)
    assert items[1][0] == "dir1"
    assert isinstance(items[1][1], JsonDB)
    assert items[3][0] == "file1"
    assert isinstance(items[3][1], AttrsDict)


def test_scan():
    jdb = JsonDB(testdb)
    jdb.scan()

    assert sorted(jdb.__dict__.keys()) == [
        "_store",
        "arrays",
        "dir1",
        "dir2",
        "file1",
        "file2",
        "file3",
        "path",
    ]


def test_time_validity():
    jdb = JsonDB(testdb)
    assert isinstance(jdb["dir1"].on("20220628T221955Z"), AttrsDict)

    assert jdb["dir1"].on("20220628T221955Z")["data"] == 1
    assert jdb.dir1.on("20220629T221955Z").data == 2
    # time point in between
    assert jdb["dir1"].on("20220628T233500Z")["data"] == 1
    # time point after
    assert jdb["dir1"].on("20220630T233500Z")["data"] == 2
    # time point before
    with pytest.raises(RuntimeError):
        jdb["dir1"].on("20220627T233500Z")["data"]

    # directory with no .jsonl
    with pytest.raises(RuntimeError):
        jdb["dir1"]["dir2"].on("20220627T233500Z")

    # invalid timestamp
    with pytest.raises(ValueError):
        jdb.dir1.on("20220627T2335002Z")

    # test usage of datetime object
    tstamp = datetime(2022, 6, 28, 23, 35, 00, tzinfo=timezone.utc)
    assert jdb.dir1.on(tstamp).data == 1
    assert jdb.dir1.on(tstamp, r"^file3.*", "all").data == 1

    assert jdb.dir1.on(tstamp, system="phy").data == 1
    assert jdb.dir1.on(tstamp, system="cal").data == 1


def test_mapping():
    jdb = JsonDB(testdb)

    assert isinstance(jdb.map("label"), AttrsDict)
    assert jdb.map("label")[3].data == 2
    assert jdb.map("key.label")[3].data == 2
    assert isinstance(jdb.file1.group.map("label"), AttrsDict)
    assert jdb.file1.group.map("label")["a"].data == 1
    assert jdb.file1.group.map("label")["b"].data == 2

    with pytest.raises(RuntimeError):
        jdb.map("system", unique=True)

    assert jdb.map("system", unique=False)[2].map("label")[1].data == 3
    assert jdb.map("system", unique=False)[1].map("label")[2].data == 1

    with pytest.raises(ValueError):
        jdb.map("non-existent-label")


def test_modification():
    d = AttrsDict()
    d["a"] = 1
    assert d.a == 1

    d.b = 2
    assert d["b"] == 2


def test_merging():
    d = AttrsDict({"a": 1})
    d |= {"b": 2}
    assert d == {"a": 1, "b": 2}
    assert isinstance(d, AttrsDict)
    assert hasattr(d, "a")
    assert hasattr(d, "b")

    d2 = d | {"c": 3}
    assert isinstance(d2, AttrsDict)
    assert d2 == {"a": 1, "b": 2, "c": 3}
    assert hasattr(d2, "a")
    assert hasattr(d2, "b")
    assert hasattr(d2, "c")

    jdb = JsonDB(testdb)
    j = jdb.dir1 | jdb.dir2
    assert isinstance(j, AttrsDict)
    assert sorted(j.keys()) == ["dir2", "file3", "file5", "file7", "file8"]
    assert hasattr(j, "dir2")
    assert hasattr(j, "file8")

    with pytest.raises(TypeError):
        jdb |= jdb.dir1



# Copyright (C) 2022 Luigi Pertoldi <gipert@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import os
from datetime import datetime
from getpass import getuser
from tempfile import gettempdir

from git import GitCommandError, InvalidGitRepositoryError, Repo

from .jsondb import AttrsDict, JsonDB

log = logging.getLogger(__name__)


class LegendMetadata(JsonDB):
    """LEGEND metadata.

    Class representing the LEGEND metadata repository with utilities for fast
    access.

    Parameters
    ----------
    path
        path to legend-metadata repository. If not existing, will attempt a
        git-clone through SSH. If ``None``, legend-metadata will be cloned
        in a temporary directory (see :func:`tempfile.gettempdir`).
    """

    def __init__(self, path: str | None = None) -> None:
        self._default_git_ref = "main"

        if isinstance(path, str):
            self._repo_path = path
        else:
            self._repo_path = os.getenv(
                "LEGEND_METADATA",
                os.path.join(gettempdir(), "legend-metadata-" + getuser()),
            )

        self._repo: Repo = self._init_metadata_repo()

        super().__init__(self._repo_path)

    def _init_metadata_repo(self):
        """Clone legend-metadata, if not existing, and checkout default Git ref."""
        if not os.path.exists(self._repo_path):
            os.mkdir(self._repo_path)

        repo = None
        try:
            repo = Repo(self._repo_path)
        except InvalidGitRepositoryError:
            log.info(
                f"Cloning git@github.com:legend-exp/legend-metadata in {self._repo_path}..."
            )
            repo = Repo.clone_from(
                "git@github.com:legend-exp/legend-metadata",
                self._repo_path,
                multi_options=["--recurse-submodules"],
            )
            repo.git.checkout(self._default_git_ref)

        return repo

    def checkout(self, git_ref: str) -> None:
        """Select a legend-metadata version."""
        try:
            self._repo.git.checkout(git_ref)
            self._repo.git.submodule("update", "--init")
        except GitCommandError:
            self._repo.remote().pull()
            self._repo.git.checkout(git_ref)
            self._repo.git.submodule("update", "--init")

    def reset(self) -> None:
        """Checkout legend-metadata to the default Git ref."""
        self._repo.git.checkout(self._default_git_ref)

    def channelmap(self, on: str | datetime = None) -> AttrsDict:
        """Get a LEGEND channel map.

        Aliases ``legend-metadata.hardware.configuration.channelmaps.on()`` and
        tries to merge the returned channel map with the detector database
        `legend-metadata.hardware.detectors` and the analysis channel map
        `dataprod.config.on(...).analysis`.

        Warning
        -------
        This method assumes ``legend-exp/legend-metadata`` has a certain
        layout. Might stop working if changes are made to the structure of the
        repository.

        Examples
        --------
        >>> from legendmeta import LegendMetadata
        >>> from datetime import datetime
        >>> channel = lmeta.channelmap(on=datetime.now()).V05267B
        >>> channel.geometry.mass_in_g
        2362.0
        >>> channel.analysis.usability
        'on'

        See Also
        --------
        .jsondb.JsonDB.on
        """
        if not on:
            on = datetime.now()

        chmap = self.hardware.configuration.channelmaps.on(
            on, pattern=None, system="all"
        )

        # get analysis metadata
        anamap = self.dataprod.config.on(on, pattern=None, system="all").analysis

        # get full detector db
        detdb = self.hardware.detectors
        fulldb = detdb.germanium.diodes | detdb.lar.sipms

        for det in chmap.keys():
            # find channel info in detector database and merge it into
            # channelmap item, if possible
            if det in fulldb:
                chmap[det] |= fulldb[det]
            else:
                log.debug(
                    f"Could not find detector '{det}' in hardware.detectors database"
                )

            # find channel info in analysis database and add it into channelmap
            # item under "analysis", if possible
            if det in anamap:
                chmap[det]["analysis"] = anamap[det]
            else:
                log.debug(
                    f"Could not find detector '{det}' in dataprod.config database"
                )

        return chmap



from copy import deepcopy

from legendmeta import police


def test_len_nested():
    assert police.len_nested({}) == 0
    assert police.len_nested({"a": 1}) == 1
    assert police.len_nested({"a": 1, "b": 2}) == 2
    assert police.len_nested({"a": 1, "b": {"c": 2}}) == 3
    assert police.len_nested({"a": 1, "b": {"c": 2, "d": "boh"}}) == 4


def test_validate_keys():
    template = {"a": 1, "b": 2, "c": {"d": "sick!", "e": 420}}

    assert police.validate_keys_recursive(template, template)
    assert police.validate_keys_recursive({"a": 3}, template)
    assert police.validate_keys_recursive({"a": 3, "c": {"d": -1}}, template)

    assert not police.validate_keys_recursive({}, template)
    assert not police.validate_keys_recursive({"z": 1}, template)
    assert not police.validate_keys_recursive({"z": 1, "h": {"b": 2}}, template)
    assert not police.validate_keys_recursive({"c": {"b": 2}}, template)


def test_validate_dict_schema():
    template = {"a": 0, "b": 0, "c": {"d": r"^\d*$", "e": 4.2}}

    case = {"a": 3, "b": 65, "c": {"d": "123", "e": 6.9}}

    assert police.validate_dict_schema(case, template)

    case1 = deepcopy(case)
    case1["c"]["d"] = "sick!"
    assert not police.validate_dict_schema(case1, template)

    case1 = deepcopy(case)
    case1["a"] = "sick!"
    assert not police.validate_dict_schema(case1, template)

    case1 = deepcopy(case)
    case1["a"] = 3.3
    assert not police.validate_dict_schema(case1, template)

    case1 = deepcopy(case)
    case1["z"] = 1
    assert not police.validate_dict_schema(case1, template)
    assert police.validate_dict_schema(case1, template, greedy=False)

    case1 = deepcopy(case)
    case1["c"]["z"] = 1
    assert not police.validate_dict_schema(case1, template)



# Copyright (C) 2022 Luigi Pertoldi <gipert@pm.me>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import argparse
import json
import os
import re
import sys
from importlib import resources

from .jsondb import JsonDB

templates = resources.files("legendmeta") / "templates"


def validate_legend_detector_db() -> bool:
    """Validate LEGEND detector database.

    Invoked in CLI.
    """
    parser = argparse.ArgumentParser(
        prog="validate-legend-detdb", description="Validate LEGEND detector database"
    )

    parser.add_argument("files", nargs="+", help="JSON files")

    args = parser.parse_args()

    dict_temp = {
        "bege": json.load(open(str(templates / "bege-detector.json"))),
        "ppc": json.load(open(str(templates / "ppc-detector.json"))),
        "coax": json.load(open(str(templates / "coax-detector.json"))),
        "icpc": json.load(open(str(templates / "icpc-detector.json"))),
    }

    for file in args.files:
        valid = True

        with open(file) as f:
            entry = json.load(f)

            if "type" not in entry:
                print(  # noqa: T201
                    f"ERROR: '{file}' entry does not contain 'type' key"
                )
                valid *= False
                continue

            if entry["type"] not in dict_temp:
                print(  # noqa: T201
                    f"WARNING: '{file}': no template for type '{entry['type']}' detector"
                )
                continue

            valid *= validate_dict_schema(
                entry,
                dict_temp[entry["type"]],
                greedy=False,
                typecheck=True,
                root_obj=file,
            )

        if not valid:
            sys.exit(1)


def validate_legend_channel_map() -> bool:
    """Validate list of LEGEND channel map files.

    Invoked in CLI.
    """
    parser = argparse.ArgumentParser(
        prog="validate-legend-chmaps", description="Validate LEGEND channel map files"
    )

    parser.add_argument("files", nargs="+", help="JSON channel maps files")

    args = parser.parse_args()

    dict_temp = {
        "geds": json.load(open(str(templates / "geds-channel.json"))),
        "spms": json.load(open(str(templates / "spms-channel.json"))),
    }

    for d in {os.path.dirname(f) for f in args.files}:
        db = JsonDB(d)
        valid = True

        with open(f"{d}/validity.jsonl") as f:
            for line in f.readlines():
                ts = json.loads(line)["valid_from"]
                sy = json.loads(line)["category"]
                chmap = db.on(ts, system=sy)

                for k, v in chmap.items():
                    if "system" not in v:
                        print(  # noqa: T201
                            f"ERROR: '{k}' entry does not contain 'system' key"
                        )
                        valid *= False
                        continue

                    if v["system"] not in dict_temp:
                        print(  # noqa: T201
                            f"WARNING: '{k}': no template for system '{v['system']}' entry"
                        )
                        continue

                    valid *= validate_dict_schema(
                        v,
                        dict_temp[v["system"]],
                        greedy=False,
                        typecheck=False,
                        root_obj=k,
                    )

        if not valid:
            sys.exit(1)


def validate_dict_schema(
    adict: dict, template: dict, greedy: bool = True, typecheck=True, root_obj: str = ""
) -> bool:
    """Validate the format of a dictionary based on a template.

    Prints error messages on the console. Returns false if dictionary is invalid.

    Parameters
    ----------
    adict
        dictionary to analyze.
    template
        template dictionary.
    greedy
        if false, do not fail if the analyzed dictionary contains extra keys.
    typecheck
        if true, perform type checking.
    root_obj
        key name (or path to) dictionary. Used for error printing.
    """
    if not isinstance(adict, dict) or not isinstance(template, dict):
        raise ValueError("input objects must be of type dict")

    valid: bool = True

    # make sure keys in template exist and are valid in adict
    for k, v in template.items():
        if k not in adict:
            print(f"ERROR: '{root_obj}/{k}' key not found")  # noqa: T201
            valid = False
        elif isinstance(v, dict):
            if not isinstance(adict[k], dict):
                print(f"ERROR: '{root_obj}/{k}' must be a dictionary")  # noqa: T201
                valid = False
            else:
                valid *= validate_dict_schema(
                    adict[k],
                    v,
                    greedy=greedy,
                    typecheck=typecheck,
                    root_obj=f"{root_obj}/{k}",
                )
        else:
            if typecheck and not isinstance(adict[k], type(v)):
                # do not complain if float is requested but int is given
                if isinstance(v, float) and isinstance(adict[k], int):
                    continue
                # make an exception for null (missing) fields
                if adict[k] is None:
                    continue
                print(  # noqa: T201
                    f"ERROR: value of '{root_obj}/{k}' must be {type(v)}"
                )  # noqa: T201
                valid = False
            elif isinstance(v, str) and v != "":
                if re.match(v, adict[k]) is None:
                    print(  # noqa: T201
                        f"ERROR: key '{root_obj}/{k}' does not match template regex '{v}'"
                    )
                    valid = False

    if greedy and len_nested(adict) != len_nested(template):
        print("ERROR: the dictionary contains extra keys")  # noqa: T201
        valid = False

    if greedy:
        valid *= validate_keys_recursive(adict, template)

    return valid


def validate_keys_recursive(adict: dict, template: dict) -> bool:
    """Return false if `adict` contains keys not in `template`."""
    valid = True

    # special case: adict is empty
    if len(adict) == 0 and len(template) != 0:
        valid = False

    # analyze adict
    for k, v in adict.items():
        if k not in template:
            print(f"ERROR: '{k}' key not allowed")  # noqa: T201
            valid = False
        elif isinstance(v, dict):
            valid *= validate_keys_recursive(v, template[k])

    return valid


def len_nested(d: dict) -> int:
    """Recursively count keys in a dictionary."""
    count = 0
    for v in d.values():
        count += 1
        if isinstance(v, dict):
            count += len_nested(v)

    return count



import pytest
import sqlalchemy as sql

from legendmeta import LegendMetadata, LegendSlowControlDB
from legendmeta.jsondb import AttrsDict
from legendmeta.slowcontrol import DiodeSnap

pytestmark = pytest.mark.xfail(
    run=True, reason="requires access to LEGEND slow control database"
)


@pytest.fixture(scope="session")
def scdb():
    scdb = LegendSlowControlDB()
    scdb.connect()
    return scdb


def test_connection(scdb):
    pass


def test_select(scdb):
    session = scdb.make_session()
    query = sql.select(DiodeSnap).limit(10)
    result = session.execute(query).all()
    assert len(result) == 10
    session.close()


def test_str_table_pandas(scdb):
    df = scdb.dataframe("diode_snap_last")
    assert len(df) > 0


def test_str_select_pandas(scdb):
    df = scdb.dataframe("SELECT channel, vmon FROM diode_snap LIMIT 10")
    assert len(df) == 10


def test_select_pandas(scdb):
    df = scdb.dataframe(sql.select(DiodeSnap.channel, DiodeSnap.vmon).limit(10))
    assert len(df) == 10


def test_status(scdb):
    lmeta = LegendMetadata()
    chmap = lmeta.channelmap()
    channel = chmap.V02162B
    status = scdb.status(channel)
    assert isinstance(status, AttrsDict)
    assert "vmon" and "vset" in status

    channel = chmap.S002
    status = scdb.status(channel)
    assert isinstance(status, AttrsDict)
    assert "vmon" and "vset" in status
